@@ -1,6 +1,6 @@
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
- * JFlex 1.4.1                                                             *
- * Copyright (C) 1998-2004  Gerwin Klein <lsf@jflex.de>                    *
+ * JFlex 1.4.2                                                             *
+ * Copyright (C) 1998-2008  Gerwin Klein <lsf@jflex.de>                    *
  * All rights reserved.                                                    *
  *                                                                         *
  * This program is free software; you can redistribute it and/or modify    *
@@ -31,14 +31,12 @@ import java.text.*;
  * Table compression, String packing etc. is also done here.
  *
  * @author Gerwin Klein
- * @version JFlex 1.4.1, $Revision: 2.26 $, $Date: 2004/11/07 03:12:46 $
+ * @version JFlex 1.4.2, $Revision: 358 $, $Date: 2008-05-27 16:28:29 +1000 (Tue, 27 May 2008) $
  */
 final public class Emitter {
     
   // bit masks for state attributes
   static final private int FINAL = 1;
-  static final private int PUSHBACK = 2;
-  static final private int LOOKEND = 4;
   static final private int NOLOOK = 8;
 
   static final private String date = (new SimpleDateFormat()).format(new Date());
@@ -74,13 +72,13 @@ final public class Emitter {
   /** maps actions to their switch label */
   private Hashtable actionTable = new Hashtable();
 
-  private CharClassInterval [] intervalls;
+  private CharClassInterval [] intervals;
 
   private String visibility = "public";
 
   public Emitter(File inputFile, LexParse parser, DFA dfa) throws IOException {
 
-    String name = parser.scanner.className+".java";
+    String name = getBaseName(parser.scanner.className) + ".java";
 
     File outputFile = normalize(name, inputFile);
 
@@ -95,6 +93,22 @@ final public class Emitter {
     this.skel = new Skeleton(out);
   }
 
+  /**
+   * Computes base name of the class name. Needs to take into account generics.
+   *
+   * @see LexScan#className
+   * @return the
+   */
+  public static String getBaseName(String className) {
+    int gen = className.indexOf('<');
+    if (gen < 0) {
+      return className;
+    }
+    else {
+      return className.substring(0, gen);
+    }
+  }
+  
 
   /**
    * Constructs a file in Options.getDir() or in the same directory as
@@ -102,7 +116,7 @@ final public class Emitter {
    *
    * @param name  the name (without path) of the file
    * @param path  the path where to construct the file
-   * @param input fallback location if path = <tt>null</tt>
+   * @param input fall back location if path = <tt>null</tt>
    *              (expected to be a file in the directory to write to)   
    */
   public static File normalize(String name, File input) {
@@ -166,6 +180,18 @@ final public class Emitter {
     print(i);
   }
 
+  private boolean hasGenLookAhead() {
+    return dfa.lookaheadUsed;
+  }
+  
+  private void emitLookBuffer() {
+    if (!hasGenLookAhead()) return;
+    
+    println("  /** For the backwards DFA of general lookahead statements */");
+    println("  private boolean [] zzFin = new boolean [ZZ_BUFFERSIZE+1];");
+    println();
+  }
+  
   private void emitScanError() {
     print("  private void zzScanError(int errorCode)");
     
@@ -288,15 +314,17 @@ final public class Emitter {
       println("   */"); 
     }      
     
+    String className = getBaseName(scanner.className);
+    
     println("  public static void main(String argv[]) {");
     println("    if (argv.length == 0) {");
-    println("      System.out.println(\"Usage : java "+scanner.className+" <inputfile>\");");
+    println("      System.out.println(\"Usage : java "+className+" <inputfile>\");");
     println("    }");
     println("    else {");
     println("      for (int i = 0; i < argv.length; i++) {");
-    println("        "+scanner.className+" scanner = null;");
+    println("        "+className+" scanner = null;");
     println("        try {");
-    println("          scanner = new "+scanner.className+"( new java.io.FileReader(argv[i]) );");
+    println("          scanner = new "+className+"( new java.io.FileReader(argv[i]) );");
 
     if ( scanner.standalone ) {      
       println("          while ( !scanner.zzAtEOF ) scanner."+scanner.functionName+"();");
@@ -344,16 +372,12 @@ final public class Emitter {
     println("            // store back cached positions");
     println("            zzCurrentPos  = zzCurrentPosL;");
     println("            zzMarkedPos   = zzMarkedPosL;");
-    if ( scanner.lookAheadUsed ) 
-      println("            zzPushbackPos = zzPushbackPosL;");
     println("            boolean eof = zzRefill();");
     println("            // get translated positions and possibly new buffer");
     println("            zzCurrentPosL  = zzCurrentPos;");
     println("            zzMarkedPosL   = zzMarkedPos;");
     println("            zzBufferL      = zzBuffer;");
     println("            zzEndReadL     = zzEndRead;");
-    if ( scanner.lookAheadUsed ) 
-      println("            zzPushbackPosL = zzPushbackPos;");
     println("            if (eof) {");
     println("              zzInput = YYEOF;");
     println("              break zzForAction;");  
@@ -440,41 +464,38 @@ final public class Emitter {
       
       int num = scanner.states.getNumber(name).intValue();
 
-      if (scanner.bolUsed)      
-        println("  "+visibility+" static final int "+name+" = "+2*num+";");
-      else
-        println("  "+visibility+" static final int "+name+" = "+dfa.lexState[2*num]+";");
+      println("  "+visibility+" static final int "+name+" = "+2*num+";");
     }
-    
-    if (scanner.bolUsed) {
-      println("");
-      println("  /**");
-      println("   * ZZ_LEXSTATE[l] is the state in the DFA for the lexical state l");
-      println("   * ZZ_LEXSTATE[l+1] is the state in the DFA for the lexical state l");
-      println("   *                  at the beginning of a line");
-      println("   * l is of the form l = 2*k, k a non negative integer");
-      println("   */");
-      println("  private static final int ZZ_LEXSTATE[] = { ");
+
+    // can't quite get rid of the indirection, even for non-bol lex states: 
+    // their DFA states might be the same, but their EOF actions might be different
+    // (see bug #1540228)
+    println("");
+    println("  /**");
+    println("   * ZZ_LEXSTATE[l] is the state in the DFA for the lexical state l");
+    println("   * ZZ_LEXSTATE[l+1] is the state in the DFA for the lexical state l");
+    println("   *                  at the beginning of a line");
+    println("   * l is of the form l = 2*k, k a non negative integer");
+    println("   */");
+    println("  private static final int ZZ_LEXSTATE[] = { ");
   
-      int i, j = 0;
-      print("    ");
+    int i, j = 0;
+    print("    ");
+    
+    for (i = 0; i < 2*dfa.numLexStates-1; i++) {
+      print( dfa.entryState[i], 2 );
 
-      for (i = 0; i < dfa.lexState.length-1; i++) {
-        print( dfa.lexState[i], 2 );
+      print(", ");
 
-        print(", ");
-
-        if (++j >= 16) {
-          println();
-          print("    ");
-          j = 0;
-        }
+      if (++j >= 16) {
+        println();
+        print("    ");
+        j = 0;
       }
-            
-      println( dfa.lexState[i] );
-      println("  };");
-
     }
+            
+    println( dfa.entryState[i] );
+    println("  };");
   }
 
   private void emitDynamicInit() {    
@@ -531,7 +552,7 @@ final public class Emitter {
     println("    char [] map = new char[0x10000];");
     println("    int i = 0;  /* index in packed string  */");
     println("    int j = 0;  /* index in unpacked array */");
-    println("    while (i < "+2*intervalls.length+") {");
+    println("    while (i < "+2*intervals.length+") {");
     println("      int  count = packed.charAt(i++);");
     println("      char value = packed.charAt(i++);");
     println("      do map[j++] = value; while (--count > 0);");
@@ -548,7 +569,7 @@ final public class Emitter {
     println("  /** ");
     println("   * The transition table of the DFA");
     println("   */");
-    println("  private static final int ZZ_TRANS [] = {"); //XXX
+    println("  private static final int ZZ_TRANS [] = {"); 
 
     print("    ");
     for (i = 0; i < dfa.numStates; i++) {
@@ -575,9 +596,8 @@ final public class Emitter {
   }
   
   private void emitCharMapArrayUnPacked() {
-   
+  
     CharClasses cl = parser.getCharClasses();
-    intervalls = cl.getIntervalls();
     
     println("");
     println("  /** ");
@@ -589,27 +609,21 @@ final public class Emitter {
     print("    ");
     
     int max =  cl.getMaxCharCode();
-    int i = 0;     
-    while ( i < intervalls.length && intervalls[i].start <= max ) {
-
-      int end = Math.min(intervalls[i].end, max);
-      for (int c = intervalls[i].start; c <= end; c++) {
-
-        print(colMap[intervalls[i].charClass], 2);
-
-        if (c < max) {
-          print(", ");        
-          if ( ++n >= 16 ) { 
-            println();
-            print("    ");
-            n = 0; 
-          }
+	
+    // not very efficient, but good enough for <= 255 characters
+    for (char c = 0; c <= max; c++) {
+      print(colMap[cl.getClassCode(c)],2);
+      
+      if (c < max) {
+        print(", ");        
+        if ( ++n >= 16 ) { 
+          println();
+          print("    ");
+          n = 0; 
         }
       }
-
-      i++;
     }
-
+    
     println();
     println("  };");
     println();
@@ -623,9 +637,9 @@ final public class Emitter {
       return;
     }
 
-    // ignores cl.getMaxCharCode(), emits all intervalls instead
+    // ignores cl.getMaxCharCode(), emits all intervals instead
 
-    intervalls = cl.getIntervalls();
+    intervals = cl.getIntervals();
     
     println("");
     println("  /** ");
@@ -636,26 +650,34 @@ final public class Emitter {
     int n = 0;  // numbers of entries in current line    
     print("    \"");
     
-    int i = 0; 
-    while ( i < intervalls.length-1 ) {
-      int count = intervalls[i].end-intervalls[i].start+1;
-      int value = colMap[intervalls[i].charClass];
-      
+    int i = 0;
+    int count, value;
+    while ( i < intervals.length ) {
+      count = intervals[i].end-intervals[i].start+1;
+      value = colMap[intervals[i].charClass];
+
+      // count could be >= 0x10000
+      while (count > 0xFFFF) {
+        printUC(0xFFFF);
+        printUC(value);
+        count -= 0xFFFF;
+        n++;       
+      }
+        
       printUC(count);
       printUC(value);
 
-      if ( ++n >= 10 ) { 
-        println("\"+");
-        print("    \"");
-        n = 0; 
+      if (i < intervals.length-1) {
+        if ( ++n >= 10 ) { 
+          println("\"+");
+          print("    \"");
+          n = 0;
+        }
       }
 
       i++;
     }
-
-    printUC(intervalls[i].end-intervalls[i].start+1);
-    printUC(colMap[intervalls[i].charClass]);
-
+      
     println("\";");
     println();
 
@@ -713,15 +735,11 @@ final public class Emitter {
     int count = 1;
     int value = 0; 
     if ( dfa.isFinal[0]    ) value = FINAL;
-    if ( dfa.isPushback[0] ) value|= PUSHBACK;
-    if ( dfa.isLookEnd[0]  ) value|= LOOKEND;
     if ( !isTransition[0]  ) value|= NOLOOK;
        
     for (int i = 1;  i < dfa.numStates; i++) {      
       int attribute = 0;      
       if ( dfa.isFinal[i]    ) attribute = FINAL;
-      if ( dfa.isPushback[i] ) attribute|= PUSHBACK;
-      if ( dfa.isLookEnd[i]  ) attribute|= LOOKEND;
       if ( !isTransition[i]  ) attribute|= NOLOOK;
 
       if (value == attribute) {
@@ -755,21 +773,40 @@ final public class Emitter {
   }
 
   private void emitConstructorDecl() {
+    emitConstructorDecl(true);
+    
+    if ((scanner.standalone || scanner.debugOption) && 
+        scanner.ctorArgs.size() > 0) {
+      Out.warning(ErrorMessages.get(ErrorMessages.CTOR_DEBUG));
+      println();
+      emitConstructorDecl(false);
+    }
+  }
+  
+  private void emitConstructorDecl(boolean printCtorArgs) {
+ 
+    String warn = 
+        "// WARNING: this is a default constructor for " +
+        "debug/standalone only. Has no custom parameters or init code.";
+    
+    if (!printCtorArgs) println(warn); 
     
     print("  ");
 
     if ( scanner.isPublic ) print("public ");   
-    print( scanner.className );      
-    print("(java.io.Reader in)");
+    print( getBaseName(scanner.className) );      
+    print("(java.io.Reader in");
+    if (printCtorArgs) emitCtorArgs();
+    print(")");
     
-    if ( scanner.initThrow != null ) {
+    if ( scanner.initThrow != null && printCtorArgs) {
       print(" throws ");
       print( scanner.initThrow );
     }
     
     println(" {");
 
-    if ( scanner.initCode != null ) {
+    if ( scanner.initCode != null && printCtorArgs) {
       print("  ");
       print( scanner.initCode );
     }
@@ -786,22 +823,39 @@ final public class Emitter {
     println("   *");
     println("   * @param   in  the java.io.Inputstream to read input from.");
     println("   */");
-
+    if (!printCtorArgs) println(warn);
+    
     print("  ");
     if ( scanner.isPublic ) print("public ");    
-    print( scanner.className );      
-    print("(java.io.InputStream in)");
+    print( getBaseName(scanner.className) );      
+    print("(java.io.InputStream in");
+    if (printCtorArgs) emitCtorArgs();
+    print(")");
     
-    if ( scanner.initThrow != null ) {
+    if ( scanner.initThrow != null && printCtorArgs ) {
       print(" throws ");
       print( scanner.initThrow );
     }
     
     println(" {");    
-    println("    this(new java.io.InputStreamReader(in));");
+
+    print("    this(new java.io.InputStreamReader(in)");
+    if (printCtorArgs) {
+      for (int i=0; i < scanner.ctorArgs.size(); i++) {
+        print(", "+scanner.ctorArgs.elementAt(i));
+      }      
+    }
+    println(");");
+    
     println("  }");
   }
 
+  private void emitCtorArgs() {
+    for (int i = 0; i < scanner.ctorArgs.size(); i++) {
+      print(", "+scanner.ctorTypes.elementAt(i));
+      print(" "+scanner.ctorArgs.elementAt(i));
+    }    
+  }
 
   private void emitDoEOF() {
     if ( scanner.eofCode == null ) return;
@@ -878,11 +932,6 @@ final public class Emitter {
 
     }
 
-    if ( scanner.lookAheadUsed ) {
-      println("    int zzPushbackPosL = zzPushbackPos = -1;");
-      println("    boolean zzWasPushback;");
-    }
-
     skel.emitNext();    
         
     if ( scanner.charCount ) {
@@ -1000,13 +1049,10 @@ final public class Emitter {
       println();
     }
     else {
-      println("      zzState = zzLexicalState;");
+      println("      zzState = ZZ_LEXSTATE[zzLexicalState];");
       println();
     }
 
-    if (scanner.lookAheadUsed)
-      println("      zzWasPushback = false;");
-
     skel.emitNext();
   }
 
@@ -1019,15 +1065,7 @@ final public class Emitter {
 
     println("          int zzAttributes = zzAttrL[zzState];");
 
-    if ( scanner.lookAheadUsed ) {
-      println("          if ( (zzAttributes & "+PUSHBACK+") == "+PUSHBACK+" )");
-      println("            zzPushbackPosL = zzCurrentPosL;");
-      println();
-    }
-
     println("          if ( (zzAttributes & "+FINAL+") == "+FINAL+" ) {");
-    if ( scanner.lookAheadUsed ) 
-      println("            zzWasPushback = (zzAttributes & "+LOOKEND+") == "+LOOKEND+";");
 
     skel.emitNext();
     
@@ -1042,9 +1080,6 @@ final public class Emitter {
     println("          zzInput = zzCMapL[zzInput];");
     println();
 
-    if ( scanner.lookAheadUsed ) 
-      println("          boolean zzPushback = false;");
-      
     println("          boolean zzIsFinal = false;");
     println("          boolean zzNoLookAhead = false;");
     println();
@@ -1063,9 +1098,6 @@ final public class Emitter {
     
     println("          if ( zzIsFinal ) {");
     
-    if ( scanner.lookAheadUsed ) 
-      println("            zzWasPushback = zzPushback;");
-    
     skel.emitNext();
     
     println("            if ( zzNoLookAhead ) break zzForAction;");
@@ -1109,18 +1141,17 @@ final public class Emitter {
     e.emitInit();
 
     for (int i = 0; i < dfa.numStates; i++) {
-      int newVal; 
+      int newVal = 0; 
       if ( dfa.isFinal[i] ) {
         Action action = dfa.action[i];
-        Integer stored = (Integer) actionTable.get(action);
-        if ( stored == null ) { 
-          stored = new Integer(lastAction++);
-          actionTable.put(action, stored);
+        if (action.isEmittable()) {
+          Integer stored = (Integer) actionTable.get(action);
+          if ( stored == null ) { 
+            stored = new Integer(lastAction++);
+            actionTable.put(action, stored);
+          }
+          newVal = stored.intValue();
         }
-        newVal = stored.intValue();
-      }
-      else {
-        newVal = 0;
       }
       
       if (value == newVal) {
@@ -1150,6 +1181,40 @@ final public class Emitter {
 
       println("        case "+label+": "); 
       
+      if (action.lookAhead() == Action.FIXED_BASE) {
+        println("          // lookahead expression with fixed base length");
+        println("          zzMarkedPos = zzStartRead + "+action.getLookLength()+";");        
+      }
+      
+      if (action.lookAhead() == Action.FIXED_LOOK || 
+          action.lookAhead() == Action.FINITE_CHOICE) {
+        println("          // lookahead expression with fixed lookahead length");
+        println("          yypushback("+action.getLookLength()+");");        
+      }
+      
+      if (action.lookAhead() == Action.GENERAL_LOOK) {
+        println("          // general lookahead, find correct zzMarkedPos");
+        println("          { int zzFState = "+dfa.entryState[action.getEntryState()]+";");
+        println("            int zzFPos = zzStartRead;");
+        println("            if (zzFin.length <= zzBufferL.length) { zzFin = new boolean[zzBufferL.length+1]; }");
+        println("            boolean zzFinL[] = zzFin;");
+        println("            while (zzFState != -1 && zzFPos < zzMarkedPos) {");
+        println("              if ((zzAttrL[zzFState] & 1) == 1) { zzFinL[zzFPos] = true; } ");
+        println("              zzInput = zzBufferL[zzFPos++];");
+        println("              zzFState = zzTransL[ zzRowMapL[zzFState] + zzCMapL[zzInput] ];");
+        println("            }");
+        println("            if (zzFState != -1 && (zzAttrL[zzFState] & 1) == 1) { zzFinL[zzFPos] = true; } ");
+        println();                
+        println("            zzFState = "+dfa.entryState[action.getEntryState()+1]+";");
+        println("            zzFPos = zzMarkedPos;");
+        println("            while (!zzFinL[zzFPos] || (zzAttrL[zzFState] & 1) != 1) {");
+        println("              zzInput = zzBufferL[--zzFPos];");
+        println("              zzFState = zzTransL[ zzRowMapL[zzFState] + zzCMapL[zzInput] ];");
+        println("            };");
+        println("            zzMarkedPos = zzFPos;");
+        println("          }");
+      }
+      
       if ( scanner.debugOption ) {
         print("          System.out.println(");
         if ( scanner.lineCount )
@@ -1191,24 +1256,7 @@ final public class Emitter {
         int num = scanner.states.getNumber(name).intValue();
         Action action = eofActions.getAction(num);
 
-        // only emit code if the lex state is not redundant, so
-        // that case labels don't overlap
-        // (redundant = points to the same dfa state as another one).
-        // applies only to scanners that don't use BOL, because
-        // in BOL scanners lex states get mapped at runtime, so
-        // case labels will always be unique.
-        boolean unused = true;                
-        if (!scanner.bolUsed) {
-          Integer key = new Integer(dfa.lexState[2*num]);
-          unused = used.get(key) == null;
-          
-          if (!unused) 
-            Out.warning("Lexical states <"+name+"> and <"+used.get(key)+"> are equivalent.");
-          else
-            used.put(key,name);
-        }
-
-        if (action != null && unused) {
+        if (action != null) {
           println("            case "+name+": {");
           if ( scanner.debugOption ) {
             print("              System.out.println(");
@@ -1293,13 +1341,13 @@ final public class Emitter {
       chars = noTarget[state].characters();
   
     print("                case ");
-    print((int)chars.nextElement());
+    print(chars.nextElement());
     print(": ");
     
     while ( chars.hasMoreElements() ) {
       println();
       print("                case ");
-      print((int)chars.nextElement());
+      print(chars.nextElement());
       print(": ");
     } 
     
@@ -1307,12 +1355,6 @@ final public class Emitter {
       if ( dfa.isFinal[nextState] )
         print("zzIsFinal = true; ");
         
-      if ( dfa.isPushback[nextState] ) 
-        print("zzPushbackPosL = zzCurrentPosL; ");
-      
-      if ( dfa.isLookEnd[nextState] )
-        print("zzPushback = true; ");
-
       if ( !isTransition[nextState] )
         print("zzNoLookAhead = true; ");
         
@@ -1332,12 +1374,6 @@ final public class Emitter {
       if ( dfa.isFinal[nextState] )
         print("zzIsFinal = true; ");
         
-      if ( dfa.isPushback[nextState] ) 
-        print("zzPushbackPosL = zzCurrentPosL; ");
-
-      if ( dfa.isLookEnd[nextState] )
-        print("zzPushback = true; ");
-          
       if ( !isTransition[nextState] )
         print("zzNoLookAhead = true; ");
         
@@ -1350,11 +1386,6 @@ final public class Emitter {
       println( "break zzForAction;" );
   }
   
-  private void emitPushback() {
-    println("      if (zzWasPushback)");
-    println("        zzMarkedPos = zzPushbackPosL;");
-  }
-  
   private int getDefaultTransition(int state) {
     int max = 0;
     
@@ -1489,7 +1520,7 @@ final public class Emitter {
 
 
   /**
-   * Set up EOF code sectioin according to scanner.eofcode 
+   * Set up EOF code section according to scanner.eofcode 
    */
   private void setupEOFCode() {
     if (scanner.eofclose) {
@@ -1547,9 +1578,11 @@ final public class Emitter {
     
     if (scanner.useRowMap) 
       emitAttributes();    
-
+    
     skel.emitNext();
     
+    emitLookBuffer();
+    
     emitClassCode();
     
     skel.emitNext();
@@ -1577,9 +1610,6 @@ final public class Emitter {
     else
       emitTransitionTable();
         
-    if (scanner.lookAheadUsed) 
-      emitPushback();
-        
     skel.emitNext();
 
     emitActions();
